# Copyright 2004-2014 Tom Rothamel <pytom@bishoujo.us>
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# This file is responsible for joystick support in Ren'Py.

import os
import pygame

import renpy.display

# Do we have a joystick enabled?
enabled = False

# The old states for each axis.
old_axis_states = { }
old_hat_states = { }

def init():
    """
    Initialize the joystick system.
    """

    global enabled

    if not renpy.config.joystick:
        return

    if 'RENPY_DISABLE_JOYSTICK' in os.environ:
        return

    try:
        pygame.joystick.init()

        for i in range(0, pygame.joystick.get_count()):
            pygame.joystick.Joystick(i).init()
            enabled = True
    except:
        if renpy.config.debug:
            raise

def resolve_axis(stick, axis, state):
    # We start by checking the user-defined mappings.
    s = "Axis %d.%d %s" % (stick, axis, state)
    stickprefix = "Axis %d."%stick

    blocked = set()
    for k,v in renpy.game.preferences.joymap.iteritems():
        if s==v:
            return k
        elif v.startswith(stickprefix):
            # And keep track of other axes on the same controller.
            blocked.add(v)

    # We don't have a keymap for this joystick event, so we try defaults that
    # should work for most controllers.
    lowaxis = axis&0xffffe
    for a in (lowaxis, lowaxis+1):
        for v in ('Positive','Negative'):
            if "%s%d %s"%(stickprefix, a, v) in blocked:
                return
    if axis & 1:
        if state=='Positive':
            return 'joy_down'
        else:
            return 'joy_up'
    elif state=='Positive':
        return 'joy_right'
    else:
        return 'joy_left'

def resolve_button(controller, button):
    s = "Button %d.%d"%(controller,button)
    for k, v in renpy.game.preferences.joymap.iteritems():
        if v==s:
            return k
    # Fallback: Alternate between menu and dismiss.
    return 'joy_menu' if button&1 else 'joy_dismiss'

def event(ev):

    if not enabled:
        return ev

    if ev.type == pygame.JOYAXISMOTION:

        if (not pygame.key.get_focused()) and (not renpy.display.interface.focused):
            return None

        if ev.value >= renpy.game.preferences.joy_deadzone:
            state = "Positive"
        elif ev.value <= -renpy.game.preferences.joy_deadzone:
            state = "Negative"
        else:
            state = None

        oldstate = old_axis_states.get((ev.joy, ev.axis), None)

        if state == oldstate:
            return None

        if oldstate:
            release = resolve_axis(ev.joy, ev.axis, oldstate)
        else:
            release = None

        old_axis_states[ev.joy, ev.axis] = state

        if state:
            press = resolve_axis(ev.joy, ev.axis, state)
            string = "Axis %d.%d %s" % (ev.joy, ev.axis, state)
            if not press:
                return ev
        else:
            press = string = None

        return pygame.event.Event(renpy.display.core.JOYEVENT,
                                  press=press, release=release, string=string)

    if ev.type == pygame.JOYHATMOTION:
        if (not pygame.key.get_focused()) and (not renpy.display.interface.focused):
            return None

        oldstate = old_hat_states.get((ev.joy, ev.hat), None)
        if ev.value==oldstate:
            return None

        # On the Logitech, the hat values are:
        # Hat[0]: -1: left. 1: right.  Hat[1]: 1: up. -1: down. 0: center.
        events = []
        if oldstate:
            if oldstate[0] and oldstate[0]!=ev.value[0]:
                if oldstate[0]<0:
                    events.append(pygame.event.Event(
                        renpy.display.core.JOYEVENT,
                        press=None, release='joy_left'))
                else:
                    events.append(pygame.event.Event(
                        renpy.display.core.JOYEVENT,
                        press=None, release='joy_right'))
            if oldstate[1] and oldstate[1]!=ev.value[1]:
                if oldstate[1]<0:
                    events.append(pygame.event.Event(
                        renpy.display.core.JOYEVENT,
                        press=None, release='joy_down'))
                else:
                    events.append(pygame.event.Event(
                        renpy.display.core.JOYEVENT,
                        press=None, release='joy_up'))
        old_hat_states[(ev.joy, ev.hat)] = ev.value
        if ev.value[0]:
            if ev.value[0]<0:
                events.append(pygame.event.Event(
                    renpy.display.core.JOYEVENT,
                    press='joy_left', release=None))
            else:
                events.append(pygame.event.Event(
                    renpy.display.core.JOYEVENT,
                    press='joy_right', release=None))
        if ev.value[1]:
            if ev.value[1]<0:
                events.append(pygame.event.Event(
                    renpy.display.core.JOYEVENT,
                    press='joy_down', release=None))
            else:
                events.append(pygame.event.Event(
                    renpy.display.core.JOYEVENT,
                    press='joy_up', release=None))
        if events:
            for e in events[1:]:
                pygame.event.post(e)
            return events[0]
        return None

    if ev.type == pygame.JOYBUTTONDOWN:

        if (not pygame.key.get_focused()) and (not renpy.display.interface.focused):
            return None

        return pygame.event.Event(renpy.display.core.JOYEVENT,
                                  press=resolve_button(ev.joy, ev.button),
                                  release=None,
                                  string="Button %d.%d"%(ev.joy, ev.button),
                                  )
    if ev.type == pygame.JOYBUTTONUP:

        if (not pygame.key.get_focused()) and (not renpy.display.interface.focused):
            return None

        return pygame.event.Event(renpy.display.core.JOYEVENT,
                                  press=None,
                                  release=resolve_button(ev.joy, ev.button),
                                  string="Button %d.%d"%(ev.joy, ev.button),
                                  )

    return ev

class JoyBehavior(renpy.display.layout.Null):
    """
    This is a behavior intended for joystick calibration. If a joystick
    event occurs, this returns it as a string.
    """

    def event(self, ev, x, y, st):
        if ev.type == renpy.display.core.JOYEVENT and ev.press:
            try:
                print "Accepting",ev.string
                return ev.string
            except:
                print "Debug: No string for",ev,ev.press
                pass
        elif ev.type == pygame.JOYAXISMOTION and abs(ev.value)>=renpy.game.preferences.joy_deadzone:
            return "Axis %d.%d %s" % (ev.joy, ev.axis, 'Positive' if ev.value>0 else 'Negative')
            

